'''RALEIGH core solver.

    Contains core eigensolver routine implemented in terms of an abstract
    Vectors class.

    Direct use is for advanced users only - start by trying user-friendly
    interfaces in raleigh/interfaces.
'''
