'''Example scripts.

core_solver.py:
    Basic core solver usage demonstration.

sparse_evp.py:
    Computes several eigenvalues and eigenvectors of a real symmetric matrix.

truncated_svd.py:
    Truncated SVD and PCA for a randomly generated matrix.
'''
