'''RALEIGH: RAL EIGensolver for real symmetric and Hermitian problems.

core:
    Contains core eigensolver routine implemented in terms of an abstract
    Vectors class.
    Direct use is for advanced users only - start by trying user-friendly
    interfaces in ./interfaces.

algebra:
    Implementations of Vectors class (and related matrix classes) on CPU and
    GPU.

interfaces:
    User-friendly interfaces to solver.

examples:
    Example scripts demonstrating the use of interfaces for the solution of
    eigenvalue problems for sparse matrices and singular value problems for
    dense matrices.
'''
